/***************************************************************************
              as10k1.c  -  Main assembler routine
                 -------------------
    Date         : May 22, 2000
    Copyright        : (C) 2000 by Daniel Bertrand
    Email        : d.bertrand@ieee.ca
 ***************************************************************************/

/*
 * This program was changed to conform the ALSA ideas. Please,
 * bug reports and all other things should be discussed on the
 * <alsa-devel@alsa-project.org> mailing list.
 *                   Jaroslav Kysela <perex@perex.cz>
 */

/***************************************************************************
 *                                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                   *
 *                                     *
 ***************************************************************************/
#include<stdio.h>
#include<stdlib.h>
#include<fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include<string.h>
#include "types.h"

#include "as10k1.h"

char *ptralloc[MAXINCLUDES];
static int num_alloc;
int source_line_num=0,file_num=0;
int macro_line_num=0;
int fd = -1;
char *input=NULL,*output=NULL,*listing=NULL,*current_line=NULL,listtemp[60];


int dbg_opt=0;

char version[]="As10k1 assembler version " VERSION;

char help[]="\n"
"Usage: as10k1 [option...] asmfile\n"
"\n"
"The as10k1 assembler is for the emu10k1 dsp processor\n"
"found in Creative Lab's sblive series of sound cards.\n"
"\n"
"Author: Daniel Bertrand <d.bertrand@ieee.ca>\n"
"\n"
"Options:\n\n"
"     -o [emu10k1 file]     Specify an output file, default is based input,\n"
"               Substituting .asm to .emu10k1\n"
"     -d [dbg_options...]   Turn on debug messages.\n"
"        s          prints all symbols        \n"
"        g          prints defined gprs\n"
"        t          prints tram usage\n"
"        i          prints instructions\n"
"     -h            Prints this message  \n"
"     -v            Prints version number. \n"
"\n"
"This program is distributed under the GPL.\n"
"\n";

//  cli==Command-Line Interface && !Creative Labs Inc.

void parse_cli_args(int argc, char *argv[])
{
    int i,j;
    for(i=1;i<argc;i++){
        
        if(argv[i][0]=='-'){
            switch(argv[i][1]){

            case 'o'://specify an output file name
                i++;
                if((i==argc)||(argv[i][0]=='-')){
                    printf("Error -o option requires a destination file name\n");
                    exit(-1);
                }
                output=argv[i];
                break;
            case 'd': //turn on debugging messages
                j=0;
                i++;
                printf("Debug on, Displaying:");
                
                if((i==argc)||(argv[i][0]=='-')){
                    dbg_opt=dbg_opt|DBG_INSTR;//default
                    i--;
                    printf(" instructions by default\n");
                    goto next_cli_arg;
                }
                
                while(argv[i][j]!='\0'){
                        
                    switch(argv[i][j]){
                    case 's':
                        dbg_opt=dbg_opt|DBG_SYM;
                        printf(" Symbols,");
                        break;
                    case 'g':
                        dbg_opt=dbg_opt|DBG_GPR;
                        printf(" GPRs,");
                        break;
                    case 't':
                        dbg_opt=dbg_opt|DBG_TRAM;
                        printf(" TRAM,");
                        break;
                    case 'i':
                        dbg_opt=dbg_opt|DBG_INSTR;
                        printf(" Instructions,");
                        break;
                        
                    default:
                        printf("\b \n**Bad debug option.  ");
                        exit(-1);
                    }
                    j++;
                }
                printf("\b \n");
                break;
                        
            case 'l'://produce a listing file
                //printf("Will save list file to %s\n",  );
                i++;
                if((i==argc)||(argv[i][0]=='-')){
                    printf("Error -l option requires a destination file name\n");
                    exit(-1);
                }
                listing=argv[i];
                
                
                break;
                
            case 'h'://printf help message

            default:
                printf("%s",help);
                    
            case 'v':
                printf("%s\n",version);
                exit(0);
            }
            
                
        }else{
            if(input==NULL)
                input=argv[i];
            else{
                printf("Error, only one input file can be specified");  
                as_exit("");
                
            }

        }
next_cli_arg:
    ;
    }
}

int main( int argc, char *argv[] )
{
    int i;
    char filename[FILENAME_MAX];
    extern int ip;
    u32 val;
    
    parse_cli_args(argc,argv);
    if(input==NULL){
        printf("Error, an input file must be specified\n");
        exit(-1);
    }

    //init symbol list:
    INIT_LIST_HEAD(&sym_head);

    asm_open(input); //opens the source file and starts parsing it.
    
    if(output==NULL){
        strcpy(filename, input);
        strtok(filename,".\0");
        strcat(filename, ".emu10k1");
        output = filename;
    }
    
    if((fd = open(output, O_WRONLY|O_CREAT|O_TRUNC, 0644)) < 0){
        printf("\nError writing to file %s\n",argv[1]); 
        as_exit("error");
    }
    
    /*create header*/
    header();

    /*output number of instructions*/
    val = __cpu_to_le32(ip);
    write(fd,&val,sizeof(u16));

    /* write binary code */
    for (i = 0; i < ip * 2; i++) {
        val = __cpu_to_le32(dsp_code[i]);
        write(fd,&val,sizeof(u32));
        //for (j = 3; j >= 0; j--)
        //fprintf(fp, "%c", ((u8 *) dsp_code)[i * 4 + j]);
    }

    fdatasync(fd);
    close(fd);
    for(i=0;i<num_alloc;i++) //free mem, is this necessary, or will the kernel free it automatically?
        free(ptralloc[i]);

    return 0;  //that's it were done
}

/*this function is called to open a asm file and parse it using the parse function.
this function is called by the main function and also by the parse function
when it encounters an "INCLUDE" directive.
*/

void asm_open(char *name)
{

    int fd,i;
    int done=0;
    char string[MAX_LINE_LENGTH];
    struct stat st;
    char *next;
    int backup_line_num=source_line_num;
    int backup_file_num=file_num;

    if( (include_depth++) > max_depth){
        printf("Error: maximum recursive include depth(%d) exceeded\n",max_depth);
        as_exit("");
    }
            
    buff[num_alloc].name=name;
    source_line_num=0;
    file_num=num_alloc;

    //open the file
    if ((unsigned) (fd = open(name, O_RDONLY)) > 255){
        as_exit("error opening input file\n");
    }
    //get it's stats
    if (fstat(fd, &st) < 0){
        printf("Error occured attempting to stat %s\n", name);
        as_exit("");
    }

    if(( ptralloc[num_alloc]=(char *) malloc(st.st_size+2) )== 0){
        printf("error allocating memory for file %s\n",name);
        close(fd);
        as_exit("");
    }else{
        buff[num_alloc].mem_start=ptralloc[num_alloc];
    }

    i=num_alloc;
    num_alloc++;
    buff[i].mem_end = buff[i].mem_start+st.st_size;
    read(fd, buff[i].mem_start, st.st_size);
    close(fd);

#ifdef DEBUG
    printf("File %s opened:\n",name);
#endif

    //get each line and parse it:
    current_line=buff[i].mem_start;
    source_line_num=1;
    next=current_line;
    while(next!=buff[i].mem_end){
        while((*next!= '\n') && (next!=buff[i].mem_end) )
            next++;
        listtemp[0]='\0';
        *next='\0';

#ifdef DEBUG
        printf("%s\n",current_line);
#endif
        if(strlen(current_line)>MAX_LINE_LENGTH)
            as_exit("Parse error: Line exceeds allowable limit");
        strcpy(&string[0],current_line);

        done = parse(string,current_line);

        *next='\n';
        
        if(done==1)
            goto done;
        
        if(next!=buff[i].mem_end){
            source_line_num++;
            next++;
        }

        current_line=next;
    }

    if(done==0)
        printf("warning no END directive at end of file %s\n",name);

done:
    source_line_num=backup_line_num;
    file_num=backup_file_num;
    include_depth--;
#ifdef DEBUG    
    printf("File %s closed:\n",name);
#endif
    return;
}

void as_exit(const char *message)
{
    int i;

    if(macro_line_num!=0)
        fprintf(stderr, "** Error while expanding macro at line %d\n",macro_line_num);

    if(source_line_num!=0)
        fprintf(stderr, "** %s.\n** line number %d:\n %s\nIn file: %s\n", message, source_line_num,current_line,buff[file_num].name);
    else
        fprintf(stderr, "** Error with file:%s\n",buff[file_num].name);
    for(i=num_alloc-1;i>=0;i--)
        free(ptralloc[i]);

    exit(1);
}

void output_tram_line(struct list_head *line_head, int type)
{
    struct tram *tram_sym;
    struct list_head *entry;

    list_for_each(entry, line_head ){
        tram_sym=list_entry(entry,struct tram,tram);
        if(tram_sym->type==type){
            //printf("table read:%s,%x\n",tram_sym->data.name,tram_sym->data.address);
            tram_sym->data.address-=TRAM_ADDR_BASE;
            write(fd,&(tram_sym->data.address),sizeof(u8));
            write(fd,&(tram_sym->data.value),sizeof(u32));
        }
    }
}


//creates output header
void header(void)
{
    int i;
    struct sym *sym;

    extern struct list_head sym_head;
    struct list_head *entry;
    
    //patch signature
       //1234567890123456
    write(fd, "EMU10K1 FX8010 1\n", 11);
    
    //patchname
    write(fd,patch_name,PATCH_NAME_SIZE);

    write(fd,&gpr_input_count,sizeof(u8));
    //write ins/outs

    list_for_each(entry,&sym_head){
        sym=list_entry(entry,struct sym,list);
        if(sym->type==GPR_TYPE_INPUT){
            sym->data.address-=GPR_BASE;
            write(fd,&(sym->data.address),sizeof(u8));
            sym->data.address++;
            write(fd,&(sym->data.address),sizeof(u8));
        }
     }

    /* dynamic gprs */
    write(fd,&gpr_dynamic_count,sizeof(u8));
    list_for_each(entry,&sym_head){
        sym=list_entry(entry,struct sym,list);
        if(sym->type==GPR_TYPE_DYNAMIC) {
            sym->data.address-=GPR_BASE;
            write(fd,&(sym->data.address),sizeof(u8));
        }
    }

    /* static gprs */
    write(fd,&gpr_static_count,sizeof(u8));

    list_for_each(entry,&sym_head){
        sym=list_entry(entry,struct sym,list);
        if(sym->type==GPR_TYPE_STATIC){
            u32 value;
            sym->data.address-=GPR_BASE;
            write(fd,&(sym->data.address),sizeof(u8));
            write(fd,&(sym->data.value),sizeof(u32));
        }
    }

    /* control gprs */
    write(fd,&gpr_control_count,sizeof(u8));

    list_for_each(entry,&sym_head){
        sym=list_entry(entry,struct sym,list);
        if(sym->type==GPR_TYPE_CONTROL){
            u32 value;
            sym->data.address-=GPR_BASE;
            write(fd,&(sym->data.address),sizeof(u8));
            value = __cpu_to_le32(sym->data.value);         
            write(fd,&value,sizeof(u32));
            value = __cpu_to_le32(((struct control *)sym)->min);
            write(fd,&value,sizeof(u32));
            value = __cpu_to_le32(((struct control *)sym)->max);
            write(fd,&value,sizeof(u32));
            write(fd,&(sym->data.name),MAX_SYM_LEN);
        }
    }
    
    /*constant GPRs*/
    write(fd,&gpr_constant_count,sizeof(u8));

    list_for_each(entry,&sym_head){
        sym=list_entry(entry,struct sym,list);
        if(sym->type==GPR_TYPE_CONSTANT){
            sym->data.address-=GPR_BASE;
            write(fd,&(sym->data.address),sizeof(u8));
            write(fd,&(sym->data.value),sizeof(u32));
        }
    }

    /*lookup-tables*/
    write(fd,&tram_table_count,sizeof(u8));
    
    for(i=0;i<tram_table_count;i++){
        u32 value;
        value = __cpu_to_le32(tram_lookup[i].size);
        write(fd,&value,sizeof(u32));
        // read lines
        write(fd,&(tram_lookup[i].read),sizeof(u8));
        output_tram_line(&(tram_lookup[i].tram),TYPE_TRAM_ADDR_READ);
        //write lines
        write(fd,&(tram_lookup[i].write),sizeof(u8));
        output_tram_line(&(tram_lookup[i].tram),TYPE_TRAM_ADDR_WRITE);
    }

    /*Delay Lines*/
    write(fd,&tram_delay_count,sizeof(u8));
    for(i=0;i<tram_delay_count;i++){
        write(fd,&(tram_delay[i].size),sizeof(u32));
        // read lines
        write(fd,&(tram_delay[i].read),sizeof(u8));
        output_tram_line(&(tram_delay[i].tram),TYPE_TRAM_ADDR_READ);
        //write lines
        write(fd,&(tram_delay[i].write),sizeof(u8));
        output_tram_line(&(tram_delay[i].tram),TYPE_TRAM_ADDR_WRITE);
    }
}

